#!/usr/bin/perl -w
#===============================================================================
#    Copyright (C) 2020-2021
#    AUTHORS: Serhiy Bolkun (Serhiy.Bolkun@slub-dresden.de)
#             Andreas Romeyke (Andreas.Romeyke@slub-dresden.de)
#             Jens Steidl (Jens.Steidl@slub-dresden.de)
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program. If not, see <https://www.gnu.org/licenses/>
#===============================================================================

use strict;
use warnings;

use Test::More tests => 107;
use Test::Output;
use Test::Exception;
use Test::Differences;
use Test::File;
use Path::Tiny;
use File::Copy::Recursive qw(dircopy);
use File::Path qw(rmtree);
# *** Read before running tests ***
# Usage: perl -I lib/ t/slubsipbuilderbagit.t 
# Note: While running tests of "SLUB_SIP_Builder", path "tmp/bagit" will be created.
# Please start test cases in the project directory to run all tests properly.
# Tested on Win10 and Debian 10

sub workaroundBagInfoTXT($$$){
    my $file_path = $_[0];
    my $key = $_[1];
    my $value = $_[2];

    # read a data file into an array
    open my $handle, '<', $file_path;
    chomp(my @lines = <$handle>);
    close $handle;
    # grep and change certain array element
    foreach my $zeile(@lines) {
        my @keyvalue = split(/: /, $zeile, 2);    # split on first :
        if($keyvalue[0] eq $key){
            $zeile = "$key: $value";
        }
    }
    # save array back to file
    open my $fh, '>', $file_path or die "Cannot open $file_path: $!";
    foreach (@lines){
        print $fh "$_\n"; # Print each entry in our array to the file
    }
    close $fh;
}

sub workaroundTagmanifest($$$){
    my $file_path = $_[0];
    my $key = $_[1];
    my $value = $_[2];

    # read a data file into an array
    open my $handle, '<', $file_path;
    chomp(my @lines = <$handle>);
    close $handle;
    # grep and change certain array element
    foreach my $zeile(@lines) {
        my @keyvalue = split(/  /, $zeile, 2);    # split
        if($keyvalue[1] eq $value){
            $zeile = "$key  $value";
        }
    }
    # save array back to file
    open my $fh, '>', $file_path or die "Cannot open $file_path: $!";
    foreach (@lines){
        print $fh "$_\n"; # Print each entry in our array to the file
    }
    close $fh;
}

sub cmpFilesByIgnoringCRLF_LF($$$){
    my $file_path1 = $_[0];
    my $file_path2 = $_[1];
    my $testname = $_[2];
    my @context = caller;
    my $content1 = path($file_path1)->slurp;
    my $content2 = path($file_path2)->slurp;
    $content1=~s/\r(?=\n)//g;
    $content2=~s/\r(?=\n)//g;
    unified_diff;
    eq_or_diff_data( $content1, $content2, $testname, { context=> 3, }) || diag "testcase at $context[1] line $context[2], with file1=$file_path1 file2=$file_path2\n\n";
    return;
}

sub cmpBinaries($$$) {
    my $file_path1 = $_[0];
    my $file_path2 = $_[1];
    my $testname = $_[2];
    my @context = caller;
    my $diag_cb = sub {
        diag "testcase at $context[1] line $context[2], with file1=$file_path1 file2=$file_path2, ".join(" ", @_)."\n\n";
    };
    my $content1;
    my $content2;
    eval {
        $content1 = path($file_path1)->slurp;
        $content2 = path($file_path2)->slurp;
    };
    if ($@) {
        fail $testname;
        &$diag_cb($@);
    } else {
        if ( $content1 eq $content2 ) { pass $testname;}
        else {
            fail $testname;
            &$diag_cb();
        }
    }

    return;
}

## prepare
BEGIN {
    use Path::Tiny;
    push @INC, Path::Tiny::path(__FILE__)->parent->parent->path("bin")->absolute->stringify;
    #require "slubsipbuilderbagit.pl";
    #$INC{'SLUB/LZA/SIPBuilderBagIt.pm'} = 1; # needed because inlined module
}
my $exportToArchiveDate = "2020-08-05_10-26-11";
my $ppn = "457035137";
my $K10plus = "1655506501";
my $noppn = "";
my $marc21;
my $marc21_2;
my @addBagInfo = (
    {'SLUBArchiv-sipVersion' => 'v2020.1'},
    {'SLUBArchiv-externalWorkflow' => 'kitodo'},
    {'SLUBArchiv-externalId' => '10008'},
    {'SLUBArchiv-externalIsilId' => 'DE-14'},
    {'SLUBArchiv-exportToArchiveDate' => '2020-08-05T10:26:11'},
    {'SLUBArchiv-hasConservationReason' => 'false'},
    {'SLUBArchiv-archivalValueDescription' => 'Gesetzlicher Auftrag'},
    {'SLUBArchiv-rightsVersion' => '1.0'}
);
# input
my $input_path = path(__FILE__)->parent->parent->child('export_dir_kitodo')->child('bagit');
# output
my $output_path = path(__FILE__)->parent->parent->child('tmp')->child('bagit');
my $tmp_path = path(__FILE__)->parent->parent->child('tmp');
# delete bagits from previous session
if(-e $tmp_path . '/bagit'){
    rmtree($tmp_path . '/bagit');
}
if(! -d $output_path->parent) { mkdir $output_path->parent or die "Error creating directory: tmp"; }
if(! -d $output_path) { mkdir $output_path or die "Error creating directory: tmp/bagit"; }

#~ structure of input output folders for tests
#~ .
#~ ├── export_dir_kitodo
#~ │   └─── bagit
#~ │    
#~ └── tmp
#~       └── bagit

my $ua = LWP::UserAgent->new;
my $request = HTTP::Request->new('GET' => "https://sru.bsz-bw.de/swbf");
my $response = $ua->request($request);
my $useragent_obj = LWP::UserAgent->new;
$useragent_obj->agent("MyApp/0.1 ");
$useragent_obj->timeout(3600); #1h

### tests
BEGIN { use_ok("SLUB::LZA::SIPBuilderBagIt"); }
#
SKIP: {
    skip "No response from server https://sru.bsz-bw.de/swb", 1 unless ! $response->is_error;
    like(SLUB::LZA::SIPBuilderBagIt::get_marc21_from("https://sru.bsz-bw.de/swb", $ppn, "pica.swn", "marcxmlvbos"), qr//, "get_marc21_from()");
};
#
ok( $marc21 = SLUB::LZA::SIPBuilderBagIt::get_marc21_from_catalogue($ppn), "get_marc21_from_catalogue(), valid");
like( $marc21, qr/<datafield tag="/, "get_marc21_from_catalogue(), valid");
ok( $marc21_2 = SLUB::LZA::SIPBuilderBagIt::get_marc21_from_catalogue($K10plus), "get_marc21_from_catalogue() K10plus, valid");
like( $marc21_2, qr/<datafield tag="/, "get_marc21_from_catalogue() K10plus, valid");
throws_ok{ SLUB::LZA::SIPBuilderBagIt::get_marc21_from_catalogue("");    } qr/ppn not defined/, "get_marc21_from_catalogue(), invalid";
throws_ok{ SLUB::LZA::SIPBuilderBagIt::get_marc21_from_catalogue(undef); } qr/ppn not defined/, "get_marc21_from_catalogue(), invalid";
#
is( SLUB::LZA::SIPBuilderBagIt::get_marc21_text_node($marc21, { tag => "245", code => "a", key => "Title" }), "Der Fichtelberg", "get_marc21_text_node(), valid");
#
my $marc_xml_unpatched = << "MARCXML";
<?xml version="1.0"?>
<record xmlns="http://www.loc.gov/MARC21/slim">
</record>
MARCXML
my $marc_doc = XML::LibXML->load_xml(string=>$marc_xml_unpatched);
ok( SLUB::LZA::SIPBuilderBagIt::add_marc21_schema_location($marc_doc), "add_marc21_schema_location(), valid");
like( $marc_doc->serialize(), qr/xsi:schemaLocation/, "add_marc21_schema_location(), valid");
throws_ok{ SLUB::LZA::SIPBuilderBagIt::add_marc21_schema_location(0); } qr/marc xml document required/, "add_marc21_schema_location(), invalid";
throws_ok{ SLUB::LZA::SIPBuilderBagIt::add_marc21_schema_location(""); } qr/marc xml document required/, "add_marc21_schema_location(), invalid";
throws_ok{ SLUB::LZA::SIPBuilderBagIt::add_marc21_schema_location(undef); } qr/marc xml document required/, "add_marc21_schema_location(), invalid";
#
my @baginfo_got = @addBagInfo;
my @baginfo_expected = @addBagInfo;
push @baginfo_expected, (
    { "External-Identifier" => "PPN:457035137" },
    { "Title" => "Der Fichtelberg" },
    { "Title" => "Berg der unbekannten Rekorde" },
    { "Title" => "ein Film von Dirk Schneider" },
    { "Title" => "Der Osten - entdecke wo du lebst" }
);
ok( SLUB::LZA::SIPBuilderBagIt::add_metadata(\@baginfo_got, $ppn, $noppn, $marc21), "add_metadata(), valid");
is_deeply( \@baginfo_got, \@baginfo_expected, "add_metadata(), valid");
#
is( SLUB::LZA::SIPBuilderBagIt::checkForTitle(\@baginfo_expected), 1, "checkForTitle(), valid");
is( SLUB::LZA::SIPBuilderBagIt::checkForTitle(\@addBagInfo),       0, "checkForTitle(), valid"); # no 'Title'
#
is(SLUB::LZA::SIPBuilderBagIt::generateBagName($exportToArchiveDate, $ppn, $noppn), "PPN-457035137_2020-08-05_10-26-11", "generateBagName()");
#
my $output_dir = $output_path->child("dir");
SLUB::LZA::SIPBuilderBagIt::createDir($output_dir);
is(-d $output_dir, 1, "createDir(\$path) path is directory");
is(-e $output_dir, 1, "createDir(\$path) directory exists in path $output_dir");
#
my $output_metaPath = $output_path->child("dir")->child("meta");
my $output_marc21FilePath = $output_metaPath->child("marc21.xml");  # auto generated in copyFilesToMeta()
my $output_rightsFilePath = $output_metaPath->child("rights.xml");
my $output_lidoFilePath = $output_metaPath->child("lido.xml");
my $output_bubbleFilePath = $output_metaPath->child("bubble.xml");
my $output_sameMarc21FilePath = $output_metaPath->child("1.xml");
my $output_sameRightsFilePath = $output_metaPath->child("2.xml");
my $input_metaPath = path(__FILE__)->parent->parent->child('export_dir_kitodo')->child('bagit')->child("meta");
my $input_marc21FilePath = $input_metaPath->child("marc21.xml");
my $input_rightsFilePath = $input_metaPath->child("rights.xml");
my $input_lidoFilePath = $input_metaPath->child("lido.xml");
my $input_bubbleFilePath = $input_metaPath->child("bubble.xml");
my @input_addMetaFile = ($input_lidoFilePath, $input_bubbleFilePath, $input_marc21FilePath, $input_rightsFilePath);
SLUB::LZA::SIPBuilderBagIt::createDir($output_metaPath);
SLUB::LZA::SIPBuilderBagIt::copyFilesToMeta($ppn, $output_metaPath, $input_rightsFilePath, \@input_addMetaFile, $marc21);
is(-f $output_marc21FilePath, 1, "copyFilesToMeta() file $output_marc21FilePath exists.");
is(-f $output_rightsFilePath, 1, "copyFilesToMeta() file $output_rightsFilePath exists.");
is(-f $output_lidoFilePath, 1, "copyFilesToMeta() file $output_lidoFilePath exists.");
is(-f $output_bubbleFilePath, 1, "copyFilesToMeta() file $output_bubbleFilePath exists.");
is(-f $output_sameMarc21FilePath, 1, "copyFilesToMeta() file $output_sameMarc21FilePath exists. Renamed self included marc21.xml");
is(-f $output_sameRightsFilePath, 1, "copyFilesToMeta() file $output_sameRightsFilePath exists. Renamed self included rights.xml");
cmpFilesByIgnoringCRLF_LF($output_rightsFilePath, $input_rightsFilePath, "copyFilesToMeta() file $input_rightsFilePath is identical to $output_rightsFilePath");
cmpFilesByIgnoringCRLF_LF($output_lidoFilePath, $input_lidoFilePath, "copyFilesToMeta() file $input_lidoFilePath is identical to $output_lidoFilePath");
cmpFilesByIgnoringCRLF_LF($output_bubbleFilePath, $input_bubbleFilePath, "copyFilesToMeta() file $input_bubbleFilePath is identical to $output_bubbleFilePath");
cmpFilesByIgnoringCRLF_LF($output_sameMarc21FilePath, $input_marc21FilePath, "copyFilesToMeta() file $input_marc21FilePath is identical to $output_sameMarc21FilePath");
cmpFilesByIgnoringCRLF_LF($output_sameRightsFilePath, $input_rightsFilePath, "copyFilesToMeta() file $input_rightsFilePath is identical to $output_sameRightsFilePath");
#
my $xsdName = "rights1.xsd";
my $input_rights = path(__FILE__)->parent->parent->child('export_dir_kitodo')->child('bagit')->child("rights");
is(SLUB::LZA::SIPBuilderBagIt::validateRightsXML($input_rights . "/Fallbeispiel-01.xml", $xsdName), 1, "validateRightsXML() Fallbeispiel-01.xml according to $xsdName");
is(SLUB::LZA::SIPBuilderBagIt::validateRightsXML($input_rights . "/Fallbeispiel-02.xml", $xsdName), 1, "validateRightsXML() Fallbeispiel-02.xml according to $xsdName");
is(SLUB::LZA::SIPBuilderBagIt::validateRightsXML($input_rights . "/Fallbeispiel-03.xml", $xsdName), 1, "validateRightsXML() Fallbeispiel-03.xml according to $xsdName");
is(SLUB::LZA::SIPBuilderBagIt::validateRightsXML($input_rights . "/Fallbeispiel-04.xml", $xsdName), 1, "validateRightsXML() Fallbeispiel-04.xml according to $xsdName");
is(SLUB::LZA::SIPBuilderBagIt::validateRightsXML($input_rights . "/Fallbeispiel-05.xml", $xsdName), 1, "validateRightsXML() Fallbeispiel-05.xml according to $xsdName");
is(SLUB::LZA::SIPBuilderBagIt::validateRightsXML($input_rights . "/Fallbeispiel-06.xml", $xsdName), 1, "validateRightsXML() Fallbeispiel-06.xml according to $xsdName");
is(SLUB::LZA::SIPBuilderBagIt::validateRightsXML($input_rights . "/Fallbeispiel-07.xml", $xsdName), 1, "validateRightsXML() Fallbeispiel-07.xml according to $xsdName");
is(SLUB::LZA::SIPBuilderBagIt::validateRightsXML($input_rights . "/Fallbeispiel-08a-undef.xml", $xsdName), 1, "validateRightsXML() Fallbeispiel-08a-undef.xml according to $xsdName");
is(SLUB::LZA::SIPBuilderBagIt::validateRightsXML($input_rights . "/Fallbeispiel-09.xml", $xsdName), 1, "validateRightsXML() Fallbeispiel-09.xml according to $xsdName");
is(SLUB::LZA::SIPBuilderBagIt::validateRightsXML($input_rights . "/Fallbeispiel-10.xml", $xsdName), 1, "validateRightsXML() Fallbeispiel-10.xml according to $xsdName");
is(SLUB::LZA::SIPBuilderBagIt::validateRightsXML($input_rights . "/Fallbeispiel-11.xml", $xsdName), 1, "validateRightsXML() Fallbeispiel-11.xml according to $xsdName");
is(SLUB::LZA::SIPBuilderBagIt::validateRightsXML($input_rights . "/Fallbeispiel-12.xml", $xsdName), 1, "validateRightsXML() Fallbeispiel-12.xml according to $xsdName");
is(SLUB::LZA::SIPBuilderBagIt::validateRightsXML($input_rights . "/Fallbeispiel-13.xml", $xsdName), 1, "validateRightsXML() Fallbeispiel-13.xml according to $xsdName");
is(SLUB::LZA::SIPBuilderBagIt::validateRightsXML($input_rights . "/Fallbeispiel-14.xml", $xsdName), 1, "validateRightsXML() Fallbeispiel-14.xml according to $xsdName");
is(SLUB::LZA::SIPBuilderBagIt::validateRightsXML($input_rights . "/Fallbeispiel-15.xml", $xsdName), 1, "validateRightsXML() Fallbeispiel-15.xml according to $xsdName");
is(SLUB::LZA::SIPBuilderBagIt::validateRightsXML($input_rights . "/Fallbeispiel-16.xml", $xsdName), 1, "validateRightsXML() Fallbeispiel-16.xml according to $xsdName");
is(SLUB::LZA::SIPBuilderBagIt::validateRightsXML($input_rightsFilePath, $xsdName), 1, "validateRightsXML() meta/rights.xml according to $xsdName");
throws_ok{ SLUB::LZA::SIPBuilderBagIt::validateRightsXML($input_metaPath . "/marc21.xml", $xsdName); } qr/failed validation/, "validateRightsXML(\$rightsFilePath, \$xsdName), rights.xml is not valid according to $xsdName";
#

my $ie = 'test';
my $bag = 'test_bag';
my $valid_bag_linux = 'lin_valid_test_bag';
my $valid_bag_windows = 'win_valid_test_bag';
my @addMetaFile = ();
my $ie_input_path = $input_path . '/' . $ie;
my $bag_output_path = $output_path . '/' . $bag;
my $data_output_path = $bag_output_path . '/data'; 
my $meta_output_path = $bag_output_path . '/meta';
my $valid_bag_linux_path = $input_path . '/' . $valid_bag_linux;
my $valid_bag_windows_path=$input_path . '/' . $valid_bag_windows;
SLUB::LZA::SIPBuilderBagIt::createDir($bag_output_path);
SLUB::LZA::SIPBuilderBagIt::createDir($data_output_path);
File::Copy::Recursive::dircopy($ie_input_path, $data_output_path) or die "Could not perform dircopy() of $ie_input_path to autogenerated $data_output_path: $!";
SLUB::LZA::SIPBuilderBagIt::createDir($meta_output_path);
SLUB::LZA::SIPBuilderBagIt::copyFilesToMeta($ppn, $meta_output_path, $input_rightsFilePath, \@addMetaFile, $marc21);

ok(SLUB::LZA::SIPBuilderBagIt::create_slub_bagit($bag_output_path, \@addBagInfo), "create_slub_bagit()");
throws_ok{ SLUB::LZA::SIPBuilderBagIt::create_slub_bagit(undef, \@addBagInfo); } qr/bagPath is not defined/, "create_slub_bagit(\$bagPath, \$refAddBagInfo), bagPath undef";
throws_ok{ SLUB::LZA::SIPBuilderBagIt::create_slub_bagit($bag_output_path, undef); } qr/array of hashes not defined for bag-info.txt/, "create_slub_bagit(\$bagPath, \$refAddBagInfo), refAddBagInfo undef";
throws_ok{ SLUB::LZA::SIPBuilderBagIt::create_slub_bagit("", \@addBagInfo); } qr/path .* is not directory/, "create_slub_bagit(\$bagPath, \$refAddBagInfo), bagPath is not a directory";

# workaround bag-info.txt, tagmanifest-md5.txt, tagmanifest-sha512.txtq
workaroundBagInfoTXT($bag_output_path . '/bag-info.txt', 'Bagging-Date', 'xxxx-xx-xx');
workaroundTagmanifest($bag_output_path . '/tagmanifest-md5.txt', 'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx', 'bag-info.txt');
workaroundTagmanifest($bag_output_path . '/tagmanifest-sha512.txt', 'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx', 'bag-info.txt');

# equalFilesByIgnoring() our testroutine to check paths
cmpBinaries($bag_output_path . '/data/00000001.tif', $valid_bag_linux_path . '/data/00000001.tif', "create_slub_bagit() $bag: validate data/00000001.tif");
cmpFilesByIgnoringCRLF_LF($bag_output_path . '/meta/marc21.xml', $valid_bag_linux_path . '/meta/marc21.xml', "create_slub_bagit() $bag: validate meta/marc21.xml");

cmpFilesByIgnoringCRLF_LF($bag_output_path . '/meta/rights.xml', $valid_bag_linux_path . '/meta/rights.xml', "create_slub_bagit() $bag: validate meta/rights.xml");
cmpFilesByIgnoringCRLF_LF($bag_output_path . '/bag-info.txt', $valid_bag_linux_path . '/bag-info.txt', "create_slub_bagit() $bag: validate bag-info.txt"); # workaround
cmpFilesByIgnoringCRLF_LF($bag_output_path . '/bagit.txt', $valid_bag_linux_path . '/bagit.txt', "create_slub_bagit() $bag: validate bagit.txt");
#
cmpBinaries($bag_output_path . '/data/00000001.tif', $valid_bag_windows_path . '/data/00000001.tif', "create_slub_bagit() $bag: validate data/00000001.tif");
cmpFilesByIgnoringCRLF_LF($bag_output_path . '/meta/marc21.xml', $valid_bag_windows_path . '/meta/marc21.xml', "create_slub_bagit() $bag: validate meta/marc21.xml");

cmpFilesByIgnoringCRLF_LF($bag_output_path . '/meta/rights.xml', $valid_bag_windows_path . '/meta/rights.xml', "create_slub_bagit() $bag: validate meta/rights.xml");
cmpFilesByIgnoringCRLF_LF($bag_output_path . '/bag-info.txt', $valid_bag_windows_path . '/bag-info.txt', "create_slub_bagit() $bag: validate bag-info.txt"); # workaround
cmpFilesByIgnoringCRLF_LF($bag_output_path . '/bagit.txt', $valid_bag_windows_path . '/bagit.txt', "create_slub_bagit() $bag: validate bagit.txt");

#
$ie = 'test2';
$bag = 'test2_bag';
$valid_bag_linux = 'lin_valid_test2_bag';
$valid_bag_windows = 'win_valid_test2_bag';
$ie_input_path = $input_path . '/' . $ie;
$bag_output_path = $output_path . '/' . $bag;
$data_output_path = $bag_output_path . '/data'; 
$meta_output_path = $bag_output_path . '/meta';
$valid_bag_linux_path = $input_path . '/' . $valid_bag_linux;
$valid_bag_windows_path = $input_path . '/' . $valid_bag_windows;
ok(SLUB::LZA::SIPBuilderBagIt::buildBagWithCopyOption($ppn, $noppn, $ie_input_path, $input_rightsFilePath, \@addMetaFile, \@addBagInfo, $bag_output_path, $data_output_path, $meta_output_path), "buildBagWithCopyOption()");

# workaround bag-info.txt, tagmanifest-md5.txt, tagmanifest-sha512.txt
workaroundBagInfoTXT($bag_output_path . '/bag-info.txt', 'Bagging-Date', 'xxxx-xx-xx');
workaroundTagmanifest($bag_output_path . '/tagmanifest-md5.txt', 'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx', 'bag-info.txt');
workaroundTagmanifest($bag_output_path . '/tagmanifest-sha512.txt', 'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx', 'bag-info.txt');

# equalFilesByIgnoring() our testroutine to check paths
cmpFilesByIgnoringCRLF_LF($bag_output_path . '/data/scans_tif/00000001.tif', $valid_bag_linux_path . '/data/scans_tif/00000001.tif', "create_slub_bagit() $bag: validate data/scans_tif/00000001.tif");
cmpFilesByIgnoringCRLF_LF($bag_output_path . '/data/scans_tif/00000002.tif', $valid_bag_linux_path . '/data/scans_tif/00000002.tif', "create_slub_bagit() $bag: validate data/scans_tif/00000002.tif");
cmpBinaries($bag_output_path . '/data/test.tiff', $valid_bag_linux_path . '/data/test.tiff', "create_slub_bagit() $bag: validate data/test.tiff");
cmpFilesByIgnoringCRLF_LF($bag_output_path . '/meta/marc21.xml', $valid_bag_linux_path . '/meta/marc21.xml', "create_slub_bagit() $bag: validate meta/marc21.xml");
cmpFilesByIgnoringCRLF_LF($bag_output_path . '/meta/rights.xml', $valid_bag_linux_path . '/meta/rights.xml', "create_slub_bagit() $bag: validate meta/rights.xml");
cmpFilesByIgnoringCRLF_LF($bag_output_path . '/bag-info.txt', $valid_bag_linux_path . '/bag-info.txt', "create_slub_bagit() $bag: validate bag-info.txt"); # workaround
cmpFilesByIgnoringCRLF_LF($bag_output_path . '/bagit.txt', $valid_bag_linux_path . '/bagit.txt', "create_slub_bagit() $bag: validate bagit.txt");
#
cmpBinaries($bag_output_path . '/data/scans_tif/00000001.tif', $valid_bag_windows_path . '/data/scans_tif/00000001.tif', "create_slub_bagit() $bag: validate data/scans_tif/00000001.tif");
cmpBinaries($bag_output_path . '/data/scans_tif/00000002.tif', $valid_bag_windows_path . '/data/scans_tif/00000002.tif', "create_slub_bagit() $bag: validate data/scans_tif/00000002.tif");
cmpBinaries($bag_output_path . '/data/test.tiff', $valid_bag_windows_path . '/data/test.tiff', "create_slub_bagit() $bag: validate data/test.tiff");
cmpFilesByIgnoringCRLF_LF($bag_output_path . '/meta/marc21.xml', $valid_bag_windows_path . '/meta/marc21.xml', "create_slub_bagit() $bag: validate meta/marc21.xml");
cmpFilesByIgnoringCRLF_LF($bag_output_path . '/meta/rights.xml', $valid_bag_windows_path . '/meta/rights.xml', "create_slub_bagit() $bag: validate meta/rights.xml");
cmpFilesByIgnoringCRLF_LF($bag_output_path . '/bag-info.txt', $valid_bag_windows_path . '/bag-info.txt', "create_slub_bagit() $bag: validate bag-info.txt"); # workaround
cmpFilesByIgnoringCRLF_LF($bag_output_path . '/bagit.txt', $valid_bag_windows_path . '/bagit.txt', "create_slub_bagit() $bag: validate bagit.txt");

$ie = 'test3';
$bag = 'test3_bag';
$valid_bag_linux = 'lin_valid_test3_bag';
$valid_bag_windows = 'win_valid_test3_bag';
$ie_input_path = $input_path . '/' . $ie;
$bag_output_path = $output_path . '/' . $bag;
$data_output_path = $bag_output_path . '/data'; 
$meta_output_path = $bag_output_path . '/meta';
$valid_bag_linux_path = $input_path.'/'.$valid_bag_linux;
$valid_bag_windows_path = $input_path . '/' . $valid_bag_windows;

my $new_ie_input_path = $tmp_path->child($ie);
File::Copy::Recursive::dircopy($ie_input_path, $new_ie_input_path) or die "Could not perform dircopy() of $ie_input_path to autogenerated $new_ie_input_path: $!";

ok(SLUB::LZA::SIPBuilderBagIt::buildBagWithMoveOption($ppn, $noppn, $new_ie_input_path, $input_rightsFilePath, \@addMetaFile, \@addBagInfo, $bag_output_path, $data_output_path, $meta_output_path), "buildBagWithMoveOption()");

# workaround bag-info.txt, tagmanifest-md5.txt, tagmanifest-sha512.txt
workaroundBagInfoTXT($bag_output_path . '/bag-info.txt', 'Bagging-Date', 'xxxx-xx-xx');
workaroundTagmanifest($bag_output_path . '/tagmanifest-md5.txt', 'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx', 'bag-info.txt');
workaroundTagmanifest($bag_output_path . '/tagmanifest-sha512.txt', 'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx', 'bag-info.txt');
cmpBinaries($bag_output_path . '/data/img/scans_tif/00000001.tif', $valid_bag_linux_path . '/data/img/scans_tif/00000001.tif', "create_slub_bagit() $bag: validate data/img/scans_tif/00000001.tif");
cmpBinaries($bag_output_path . '/data/img/scans_tif/00000002.tif', $valid_bag_linux_path . '/data/img/scans_tif/00000002.tif', "create_slub_bagit() $bag: validate data/img/scans_tif/00000002.tif");
cmpBinaries($bag_output_path . '/data/test.tiff', $valid_bag_linux_path . '/data/test.tiff', "create_slub_bagit() $bag: validate data/test.tiff");
cmpFilesByIgnoringCRLF_LF($bag_output_path . '/meta/marc21.xml', $valid_bag_linux_path . '/meta/marc21.xml', "create_slub_bagit() $bag: validate meta/marc21.xml");
cmpFilesByIgnoringCRLF_LF($bag_output_path . '/meta/rights.xml', $valid_bag_linux_path . '/meta/rights.xml', "create_slub_bagit() $bag: validate meta/rights.xml");
cmpFilesByIgnoringCRLF_LF($bag_output_path . '/bag-info.txt', $valid_bag_linux_path . '/bag-info.txt', "create_slub_bagit() $bag: validate bag-info.txt"); # workaround
cmpFilesByIgnoringCRLF_LF($bag_output_path . '/bagit.txt', $valid_bag_linux_path . '/bagit.txt', "create_slub_bagit() $bag: validate bagit.txt");
#
cmpBinaries($bag_output_path . '/data/img/scans_tif/00000001.tif', $valid_bag_windows_path . '/data/img/scans_tif/00000001.tif', "create_slub_bagit() $bag: validate data/img/scans_tif/00000001.tif");
cmpBinaries($bag_output_path . '/data/img/scans_tif/00000002.tif', $valid_bag_windows_path . '/data/img/scans_tif/00000002.tif', "create_slub_bagit() $bag: validate data/img/scans_tif/00000002.tif");
cmpBinaries($bag_output_path . '/data/test.tiff', $valid_bag_windows_path . '/data/test.tiff', "create_slub_bagit() $bag: validate data/test.tiff");
cmpFilesByIgnoringCRLF_LF($bag_output_path . '/meta/marc21.xml', $valid_bag_windows_path . '/meta/marc21.xml', "create_slub_bagit() $bag: validate meta/marc21.xml");
cmpFilesByIgnoringCRLF_LF($bag_output_path . '/meta/rights.xml', $valid_bag_windows_path . '/meta/rights.xml', "create_slub_bagit() $bag: validate meta/rights.xml");
cmpFilesByIgnoringCRLF_LF($bag_output_path . '/bag-info.txt', $valid_bag_windows_path . '/bag-info.txt', "create_slub_bagit() $bag: validate bag-info.txt"); # workaround
cmpFilesByIgnoringCRLF_LF($bag_output_path . '/bagit.txt', $valid_bag_windows_path . '/bagit.txt', "create_slub_bagit() $bag: validate bagit.txt");

$ie = 'test4';
$bag = 'test4_bag';
$valid_bag_linux = 'lin_valid_test4_bag';
$valid_bag_windows = 'win_valid_test4_bag';
$ie_input_path = $input_path . '/' . $ie;
$bag_output_path = $output_path . '/' . $bag;
$data_output_path = $bag_output_path . '/data'; 
$meta_output_path = $bag_output_path . '/meta';
$valid_bag_linux_path = $input_path . '/' . $valid_bag_linux;
$valid_bag_windows_path = $input_path . '/' . $valid_bag_windows;
$new_ie_input_path = $output_path . '/' . $ie ;
File::Copy::Recursive::dircopy($ie_input_path, $new_ie_input_path) or die "Could not perform dircopy() of $ie_input_path to autogenerated $new_ie_input_path: $!";

ok(SLUB::LZA::SIPBuilderBagIt::buildBagWithReplaceOption($ppn, $noppn, $new_ie_input_path, $input_rightsFilePath, \@addMetaFile, \@addBagInfo, $bag_output_path, $data_output_path, $meta_output_path), "buildBagWithReplaceOption()");

# workaround bag-info.txt, tagmanifest-md5.txt, tagmanifest-sha512.txt
workaroundBagInfoTXT($bag_output_path . '/bag-info.txt', 'Bagging-Date', 'xxxx-xx-xx');
workaroundTagmanifest($bag_output_path . '/tagmanifest-md5.txt', 'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx', 'bag-info.txt');
workaroundTagmanifest($bag_output_path . '/tagmanifest-sha512.txt', 'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx', 'bag-info.txt');

cmpBinaries($bag_output_path . '/data/dir_test_äöüÄÖÜß/00000001.tif', $valid_bag_linux_path . '/data/dir_test_äöüÄÖÜß/00000001.tif', "create_slub_bagit() $bag: validate data/dir_test_äöüÄÖÜß/00000001.tif");
cmpBinaries($bag_output_path . '/data/test_äöüÄÖÜß.tiff', $valid_bag_linux_path . '/data/test_äöüÄÖÜß.tiff', "create_slub_bagit() $bag: validate data/test_äöüÄÖÜß.tiff");
cmpFilesByIgnoringCRLF_LF($bag_output_path . '/meta/marc21.xml', $valid_bag_linux_path . '/meta/marc21.xml', "create_slub_bagit() $bag: validate meta/marc21.xml");
cmpFilesByIgnoringCRLF_LF($bag_output_path . '/meta/rights.xml', $valid_bag_linux_path . '/meta/rights.xml', "create_slub_bagit() $bag: validate meta/rights.xml");
cmpFilesByIgnoringCRLF_LF($bag_output_path . '/bag-info.txt', $valid_bag_linux_path . '/bag-info.txt', "create_slub_bagit() $bag: validate bag-info.txt"); # workaround
cmpFilesByIgnoringCRLF_LF($bag_output_path . '/bagit.txt', $valid_bag_linux_path . '/bagit.txt', "create_slub_bagit() $bag: validate bagit.txt");
#
cmpBinaries($bag_output_path . '/data/dir_test_äöüÄÖÜß/00000001.tif', $valid_bag_windows_path . '/data/dir_test_äöüÄÖÜß/00000001.tif', "create_slub_bagit() $bag: validate data/dir_test_äöüÄÖÜß/00000001.tif");
cmpBinaries($bag_output_path . '/data/test_äöüÄÖÜß.tiff', $valid_bag_windows_path . '/data/test_äöüÄÖÜß.tiff', "create_slub_bagit() $bag: validate data/test_äöüÄÖÜß.tiff");
cmpFilesByIgnoringCRLF_LF($bag_output_path . '/meta/marc21.xml', $valid_bag_windows_path . '/meta/marc21.xml', "create_slub_bagit() $bag: validate meta/marc21.xml");
cmpFilesByIgnoringCRLF_LF($bag_output_path . '/meta/rights.xml', $valid_bag_windows_path . '/meta/rights.xml', "create_slub_bagit() $bag: validate meta/rights.xml");
cmpFilesByIgnoringCRLF_LF($bag_output_path . '/bag-info.txt', $valid_bag_windows_path . '/bag-info.txt', "create_slub_bagit() $bag: validate bag-info.txt"); # workaround
cmpFilesByIgnoringCRLF_LF($bag_output_path . '/bagit.txt', $valid_bag_windows_path . '/bagit.txt', "create_slub_bagit() $bag: validate bagit.txt");

1;
